#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Apr  9 09:19:22 2019

@author: stan
"""
from boolparse.logic.logic import TRUE,FALSE,Var,Not,And,Or,\
                    Implication,Equivalence
class CNF:
    
    def __init__(self,formula,variables):
        self.P = "__"
        self.V = "Z"
        self.formula = formula
        self.formulaCNF = None
        self.variables = variables
        self.encode ={v:i+1 for i,v in enumerate(variables)}
        self.index = len(variables)+1
        self.dimacs = []
        
    def next_index(self):
        i = self.index
        self.index +=1
        return i
    
    def switching_var(self):
        var = Var(self.P+self.V+str(self.next_index())+self.P)
        self.encode[var.nombre]=self.index-1
        return var 
    
    def rewrite(self,formula):
        if isinstance(formula,TRUE):
            return formula
        elif isinstance(formula,FALSE):
            return formula
        elif isinstance(formula,Var):
            return formula
        elif isinstance(formula, Not):
            return Not(self.rewrite(formula.param))
        elif isinstance(formula,And):
            return And(
            self.rewrite(formula.left),self.rewrite(formula.right))
        elif isinstance(formula,Or):
            return Or(
            self.rewrite(formula.left),self.rewrite(formula.right))
        elif isinstance(formula,Implication):
            return Or(
                    Not(self.rewrite(formula.left)),
                    self.rewrite(formula.right))
        elif isinstance(formula,Equivalence):
            return And(
                    Or(
                    Not(self.rewrite(formula.left)),
                    self.rewrite(formula.right)),
                    Or(
                    Not(self.rewrite(formula.right)),
                    self.rewrite(formula.left)))
                    
    def apply_not(self,formula):
        if isinstance(formula,Not):
            if isinstance(formula.param,And):
                return Or(
                    self.apply_not(Not(formula.param.left)),
                    self.apply_not(Not(formula.param.right)))
            elif isinstance(formula.param,Or):
                return And(
                    self.apply_not(Not(formula.param.left)),
                    self.apply_not(Not(formula.param.right)))
            elif isinstance(formula.param,Not):
                return self.apply_not(formula.param.param)
            else:
                return formula
        elif isinstance(formula,And):
            return And(
                    self.apply_not(formula.left),
                    self.apply_not(formula.right))
        elif isinstance(formula,Or):
            return Or(
                    self.apply_not(formula.left),
                    self.apply_not(formula.right))
        else:
            return formula
        
    def literal(self,formula):
        if isinstance(formula,Var):
            return True
        elif isinstance(formula,Not):
            return isinstance(formula.param,Var)
        else:
            return False
        
    def distribute_ors(self,formula):
        if self.literal(formula):
            return formula
        if isinstance(formula,Or):
            if self.literal(formula.left) and self.literal(formula.right):
                return formula
            if self.literal(formula.left):
                if isinstance(formula.right,And):
                    return self.distribute(
                    formula.left,self.distribute_ors(formula.right))
                else: 
                    return Or(formula.left,
                    self.distribute_ors(formula.right))
            if self.literal(formula.right):
                if isinstance(formula.left,And):
                    return self.distribute(
                    formula.right,self.distribute_ors(formula.left))
                else: 
                    return Or(formula.right,
                    self.distribute_ors(formula.left))
            else:
                Z = self.switching_var()
                return And(
                    self.distribute(Not(Z),self.distribute_ors(formula.left)),
                    self.distribute(Z,self.distribute_ors(formula.right)))
        elif isinstance(formula,And):
            return And(
                    self.distribute_ors(formula.left),
                    self.distribute_ors(formula.right)
                    )
        else:
            return formula
            

    def distribute(self,variable,formula):
        if self.literal(formula):
            return Or(variable,formula)
        elif isinstance(formula,And):
            return And(
                    self.distribute(variable,formula.left),
                    self.distribute(variable,formula.right))
        elif isinstance(formula,Or):
            return Or(variable,formula)
        

            
    def convert(self):
        f = self.rewrite(self.formula)
        f = self.apply_not(f)
        f = self.distribute_ors(f)
        self.decode = {i:v for v,i in self.encode.items()}
        self.formulaCNF = f
        return f
    
    def to_dimacs(self,formula=None):
        formula = self.formulaCNF if formula==None else formula
        if not isinstance(formula,And):
            clausula = self.encode_clausula(formula,[])
            self.dimacs.append(clausula)
        else:
            self.to_dimacs(formula.left)
            self.to_dimacs(formula.right)
        return self.dimacs
        
    def encode_clausula(self,formula,clausula=[]):
        if self.literal(formula):
            self.agregar_literal(formula,clausula)
        else:
            self.encode_clausula(formula.left,clausula)
            self.encode_clausula(formula.right,clausula)
        return clausula
    
    def agregar_literal(self,literal,clausula):
        if isinstance(literal,Not):
            variable = literal.param 
            indice = -self.encode[variable.nombre]
        else:
            indice = self.encode[literal.nombre]
        clausula.append(indice)
    
    
    